/* 
    Zeta Dva, yet another amateur level chess engine
    Author: Srdja Matovic <srdja@matovic.de>
    Created at: 15-Jan-2011
    Updated at:
    Description: Amateur chesss programm

    Copyright (C) 2011 Srdja Matovic
    This program is distributed under the GNU General Public License.
    See file COPYING or http://www.gnu.org/licenses/

*/

#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#if defined _WIN32 || defined _WIN64
#include <windows.h>
#else
#endif

#include "book.h"
#include "types.h"
#include "eval.h"
#include "bitboard.h" // magic hashtables and bitcount from Stockfish

// global vars
static unsigned long MOVECOUNT  = 0;
static unsigned long NODECOUNT  = 0;
static bool TIMEOUT             = false;
static int alpha_cut            = 0;
static int beta_cut             = 0;
static int TT_alpha             = 0;
static int TT_beta              = 0;
static int TT_move              = 0;
static int debug_I              = 0;
static int debug_II             = 0;
int PLY                         = 0;
int SOM                         = WHITE;

// config 
static bool USE_NULL            = true;
static bool USE_LMR             = true;
static bool USE_IID             = true;
static bool USE_TT_MOVE         = true;
static bool USE_TT_BOUNDS       = true;
static bool USE_INC_HASH        = true;
static bool USE_INC_EVAL        = false;
static bool USE_LAZY_EVAL       = false;
bool USE_BOOK                   = true;
static bool USE_FUTILITY        = true;
static bool USE_RAZOR           = true; 
static bool USE_MATE_D_PRUNING  = true; 
static bool PRINT_INFO          = true;


static int max_depth            = 96;
static int max_mem_mb           = 64;
static int max_cores            = 1;
static int core                 = 0;
static bool force_mode          = false;
static bool random_mode         = false;
static bool xboard_mode         = false;

// time management
static double max_time_per_move = 0;
static double time_per_move     = 0;
static int max_moves            = 0;
static int time_seconds         = 0;
static int time_minutes         = 0;
static double time_left_opponent = 0;  
static double time_left_computer = 0;  
static int best_score           = 0;
static char time_string[128];

static clock_t start, end;
static double elapsed;


const Score  INF                = 32000;
const Score  MATESCORE          = 29000;

static U64 Hash_Stack[1024];
static Bitboard BOARD[12];
/*
00 White
01 Black
02 Pawns
03 Knights
04 Bishops
05 Rooks
06 Queens
07 Kings
08 empty/dummy
09 CR
10 hash
11 incremental score
*/

static Boardindex BINDEX[65];
Bitboard AttackTables[7][64];
static Bitboard AttackTablesTo[7][64];
static Move Lastmove = 0;

Move CounterMoves[64][64];
Move Killers[1024][2];


// functions
void move2alg(Move move, char * movec);
static void print_bitboard(Bitboard board);
static void print_board(Bitboard *board, Boardindex *bindex);
void print_movealg(Move move);
static void print_debug_movealg(Move move);
Move move_parser(char *usermove, Boardindex *bindex, int som);
int kingincheck(Bitboard *board, int som);
int checkmate(Bitboard *board, Boardindex *bindex, int som, Move lastmove , int kic);
static void setboard(char *fen);
static void setboard_epd(char *epd);
static int cmp_move_desc(const void *ap, const void *bp);
static void print_stats();
static void print_checkmate(int side);
static void print_stalemate();
static Score alphabeta(Bitboard *board, Boardindex *bindex, int som, int mdepth, int ply, Move lastmove, Score alpha, Score beta, signed nodetype, bool nullmove);

struct tt *TT;

/* ############################# */
/* ###        inits          ### */
/* ############################# */

static void init_tables() {

    int i,j,k,l;

    for(i=0; i< 64; i++) {
        for(j=0; j< 64; j++) {
            CounterMoves[i][j] = 0;
        }        
    }

    for (i=0; i<100;i++) {
        Killers[i][0] = 0;
        Killers[i][1] = 0;
    }

    // Tables for White Pawns, Black Pawns, Knights and King
    int step[7][8] =  
    {
        {0},            // empty
        {8,0},            // white pawn one step
        {7,9,0},            // white pawn capture
        {-8,0},           // black pawn one step
        {-7,-9,0},          // black pawn capture
        {17,15,10,6,-6,-10,-15,-17},    // knights
        {9,7,-7,-9,8,1,-1,-8}           // king
    };

    for (i = 0; i < 64; i++) {
        for (j = 0; j < 7; j++) {
            AttackTables[j][i] = 0ULL;
            for (k = 0; k < 8 && step[j][k] != 0; k++) {
                l = i + step[j][k];
                if (l >= 0 && l < 64 && abs((i & 7) - (l & 7)) < 3) {
                    AttackTables[j][i]   |= (1ULL << l);
                    AttackTablesTo[j][l] |= (1ULL << i); 
                }
            }
        }
    }
}

static void init_tt() {

    int i,j;
    struct tt *tt;

    if (TT != NULL)
        free(TT);
    
    // calc TT size and bits
    int temp_mem = max_mem_mb*1024*1024 / (sizeof (struct tt));
    int temp_bit = (int) log2(temp_mem);
    core = pow(2, temp_bit);

    TT = (struct tt*) malloc(core * sizeof (struct tt));

    for(i=0;i<core;i++) {
        tt = &TT[i];
        tt->hash = 0;
        tt->score = 0;
        tt->flag = FAILLOW;
        tt->depth = 0;
    }

    for(i=0; i< 64; i++) {
        for(j=0; j< 64; j++) {
            CounterMoves[i][j] = 0;
        }        
    }

    for (i=0; i<100;i++) {
        Killers[i][0] = 0;
        Killers[i][1] = 0;
    }

}

/* ############################# */
/* ###     move tools        ### */
/* ############################# */

static inline Move makemove(Square from, Square to, Square cpt, Piece pfrom, Piece pto, Piece pcpt, Cr cr) {
    return ((U64)from | (U64)to<<6 |  (U64)cpt<<12 |  (U64)pfrom<<18 | (U64)pto<<22 | (U64)pcpt<<26 | (U64)cr<<30  );  
}
static inline Move getmove(Move move) {
    return (move & 0x3FFFFFFF);  
}
static inline Move setscore(Move move, Score score) {
    return ((move & 0x3FFFFFFFF) | (U64)(score&0xFFFF)<<34);  
}
static inline Score getscore(Move move) {
    return (Score)((move>>34) &0xFFFF);  
}
static inline Square getfrom(Move move) {
    return (Square)(move & 0x3F);
}
static inline Square getto(Move move) {
    return (Square)((move>>6) & 0x3F);
}
static inline Square getcpt(Move move) {
    return (Square)((move>>12) & 0x3F);
}
static inline Piece getpfrom(Move move) {
    return (Piece)((move>>18) & 0xF);
}
static inline Piece getpto(Move move) {
    return (Piece)((move>>22) & 0xF);
}
static inline Piece getpcpt(Move move) {
    return (Piece)((move>>26) & 0xF);
}
static inline Cr getcr(Move move) {
    return (Cr)( (move>>30)&0xF);
}
static inline Move setcr(Move move, Cr cr) {
    return ( (move & 0x3FFFFFFF) |  (Move)(cr&0xF)<<30 );
}
static inline void setcastle_kingside(Bitboard *board, int som) {
    board[boardCR] |= som? 0x4 : 0x1;
}
static inline void setcastle_queenside(Bitboard *board, int som) {
    board[boardCR] |= som? 0x8 : 0x2;
}
static inline void clearcastle_kingside(Bitboard *board, int som) {
    board[boardCR] &= som? 0xB: 0xE;
}
static inline void clearcastle_queenside(Bitboard *board, int som) {
    board[boardCR] &= som? 0x7 : 0xD ;
}
static int checkcastle_kingside(Bitboard *board, int som) {
    return som ? (int)((board[boardCR]>>2)&1) : (int)((board[boardCR])&1);
}
static int checkcastle_queenside(Bitboard *board, int som) {
    return som ? (int)((board[boardCR]>>3)&1 ) : (int)((board[boardCR]>>1)&1);
}
static bool is_pdsq_move(Move move, int som) {

    return (getpfrom(move) == PAWN && abs(getto(move)-getfrom(move))==16 ) ? true: false;
}

static int get_key_piece(Piece p, int side) {

    int piece = 12;

    piece = p == PAWN?      side?  0:1    :piece;
    piece = p == KNIGHT?    side?  2:3    :piece;
    piece = p == BISHOP?    side?  4:5    :piece;
    piece = p == ROOK?      side?  6:7    :piece;
    piece = p == QUEEN?     side?  8:9    :piece;
    piece = p == KING?      side? 10:11   :piece;

    return piece;
}

/* ############################# */
/* ###      hash tools       ### */
/* ############################# */

static U64 compute_hash(Bitboard *board, int som, Move lastmove) {

   return book_gen_key(board, som, lastmove);
}

static void save_to_tt(U64 hash, Move move, Score score, signed char flag, char depth, int ply, bool qs) {

    struct tt *tete;
    
    // exit when timeout
    if ( TIMEOUT == true ) {
        return;
    }

    tete = &TT[((hash))&(core-1)];
    tete->hash = hash;
    tete->bestmove = getmove(move);
    tete->score = score;
    tete->flag = flag;
    tete->depth = depth;
    tete->ply = ply;
    tete->qs = qs;

}

struct tt *load_from_tt(U64 hash) {

    struct tt *tete;

    tete = &TT[((hash))&(core-1)];
    if (hash == tete->hash ) {
        return tete;
    }

    return NULL;

}


/* ############################# */
/* ###     domove undomove   ### */
/* ############################# */

static void donullmove(Bitboard *board, Boardindex *bindex, int som, Move lastmove) {

    board[boardHash] ^=RandomTurn[0];

    // clear en passant
    // en passant
    if (getpfrom(lastmove) == PAWN && abs(getto(lastmove)-getfrom(lastmove))==16) {
        if (som == BLACK) {
            if ( (board[BLACK]&board[PAWN]) & (SetMaskBB[getto(lastmove)+1] | SetMaskBB[getto(lastmove)-1]) & 0xFF000000 )
                board[boardHash] ^= RandomEnPassant[(getto(lastmove)-8)&7]; 
        }
        else {
            if ( (board[WHITE]&board[PAWN]) & (SetMaskBB[getto(lastmove)+1] | SetMaskBB[getto(lastmove)-1]) & 0xFF00000000)
                board[boardHash] ^= RandomEnPassant[(getto(lastmove)+8)&7]; 
        }
    }

}

static void undonullmove(Bitboard *board, Boardindex *bindex, int som, U64 hash) {

    board[boardHash] = hash;
}

static void domove(Bitboard *board, Boardindex *bindex, Move move, int som, bool doscore, Move lastmove) {

    Square from  = getfrom(move);
    Square to    = getto(move);
    Square cpt   = getcpt(move);

    Piece pfrom  = getpfrom(move);
    Piece pto    = getpto(move);
    Piece pcpt   = getpcpt(move);

    int key_piece = 0;


    // Eval increment    
    if (doscore && USE_INC_EVAL) {
        board[boardScore]-= evalIncrement(board, bindex, som, pfrom, from);
        if (pcpt >= PAWN && pcpt <= KING) {
            board[boardScore]-= evalIncrement(board, bindex, !som, pcpt, cpt);
        }
    }

    // incremental hash, clear old
    if (doscore) {
        // castle rights
        if (checkcastle_kingside(board, WHITE))
            board[boardHash] ^= RandomCastle[0];
        if (checkcastle_queenside(board, WHITE))
            board[boardHash] ^= RandomCastle[1];
        if (checkcastle_kingside(board, BLACK))
            board[boardHash] ^= RandomCastle[2];
        if (checkcastle_queenside(board, BLACK))
            board[boardHash] ^= RandomCastle[3];
        if (som == WHITE)
            board[boardHash] ^=RandomTurn[0];

        // en passant
        if (getpfrom(lastmove) == PAWN && abs(getto(lastmove)-getfrom(lastmove))==16) {
            if (som == BLACK) {
                if ( (board[BLACK]&board[PAWN]) & (SetMaskBB[getto(lastmove)+1] | SetMaskBB[getto(lastmove)-1]) & 0xFF000000 )
                    board[boardHash] ^= RandomEnPassant[(getto(lastmove)-8)&7]; 
            }
            else {
                if ( (board[WHITE]&board[PAWN]) & (SetMaskBB[getto(lastmove)+1] | SetMaskBB[getto(lastmove)-1]) & 0xFF00000000)
                    board[boardHash] ^= RandomEnPassant[(getto(lastmove)+8)&7]; 
            }
        }
    }

    // Castle move kingside move rook
    if ( (pfrom == KING) && (signed int)(to-from) == 2 ) {

        key_piece = get_key_piece(ROOK, som);
        // unset from rook
        if (doscore) {
            board[boardHash] ^= Random64[64*key_piece+from+3];
        }
        if (doscore && USE_INC_EVAL) {
            board[boardScore]-= evalIncrement(board, bindex, som, ROOK, from+3);
        }
        board[som]      &= ClearMaskBB[from+3];
        board[ROOK]     &= ClearMaskBB[from+3];
        bindex[from+3]   = PEMPTY;
        // set to rook
        board[som]      |= SetMaskBB[to-1];
        board[ROOK]     |= SetMaskBB[to-1];
        bindex[to-1]     = ROOK; 
        if (doscore) {
            board[boardHash] ^= Random64[64*key_piece+to-1];
        }
        if (doscore && USE_INC_EVAL) {
            board[boardScore]+= evalIncrement(board, bindex, som, ROOK, to-1);
        }
    }
    // Castle move queenside move rook
    if ( (pfrom == KING) && (signed int)(from-to) == 2 ) {

        key_piece = get_key_piece(ROOK, som);
        // unset from rook
        if (doscore) {
            board[boardHash] ^= Random64[64*key_piece+from-4];
        }
        if (doscore && USE_INC_EVAL) {
            board[boardScore]-= evalIncrement(board, bindex, som, ROOK, from-4);
        }
        board[som]      &= ClearMaskBB[from-4];
        board[ROOK]     &= ClearMaskBB[from-4];
        bindex[from-4]   = PEMPTY;
        // set to rook
        board[som]      |= SetMaskBB[to+1];
        board[ROOK]     |= SetMaskBB[to+1];
        bindex[to+1]     = ROOK; 
        if (doscore) {
            board[boardHash] ^= Random64[64*key_piece+to+1];
        }
        if (doscore && USE_INC_EVAL) {
            board[boardScore]+= evalIncrement(board, bindex, som, ROOK, to+1);
        }
    }

    // normal move, including king castle
    // unset capture
    if (pcpt != PEMPTY ) {
        board[!som]  &= ClearMaskBB[cpt];
        board[pcpt]  &= ClearMaskBB[cpt];
        bindex[cpt]   = PEMPTY;
    }
    // set to
    board[som]  |= SetMaskBB[to];
    board[pto]  |= SetMaskBB[to];
    bindex[to]   = pto;
    // unset from
    board[som]   &= ClearMaskBB[from];
    board[pfrom] &= ClearMaskBB[from];
    bindex[from]  = PEMPTY;


    // update castle rights, TODO: make nice
    if (pfrom == ROOK && !som && from == 0)
        clearcastle_queenside(board, WHITE);
    if (pfrom == ROOK && !som && from == 7)
        clearcastle_kingside(board, WHITE);
    if (pfrom == ROOK && som && from == 56)
        clearcastle_queenside(board, BLACK);
    if (pfrom == ROOK && som && from == 63)
        clearcastle_kingside(board, BLACK);

    if (pfrom == KING && !som && from == 4) {
        clearcastle_queenside(board, WHITE);
        clearcastle_kingside(board, WHITE);
    }
    if (pfrom == KING && som && from == 60) {
        clearcastle_queenside(board, BLACK);
        clearcastle_kingside(board, BLACK);
    }

    // incremental hash
    if (doscore) {
        key_piece = get_key_piece(pfrom, som);
        board[boardHash] ^= Random64[64*key_piece+from];
        if (pcpt >= PAWN && pcpt <= KING) {
            key_piece = get_key_piece(pcpt, !som);
            board[boardHash] ^= Random64[64*key_piece+cpt];
        }
        key_piece = get_key_piece(pto, som);
        board[boardHash] ^= Random64[64*key_piece+to];

        // incremental hash, set new castle rigths
        if (checkcastle_kingside(board, WHITE))
            board[boardHash] ^= RandomCastle[0];
        if (checkcastle_queenside(board, WHITE))
            board[boardHash] ^= RandomCastle[1];
        if (checkcastle_kingside(board, BLACK))
            board[boardHash] ^= RandomCastle[2];
        if (checkcastle_queenside(board, BLACK))
            board[boardHash] ^= RandomCastle[3];
        if (!som == WHITE)
            board[boardHash] ^=RandomTurn[0];

        // en passant
        if (getpfrom(move) == PAWN && abs(getto(move)-getfrom(move))==16) {
            if (som == WHITE) {
                if ( (board[BLACK]&board[PAWN]) & (SetMaskBB[getto(move)+1] | SetMaskBB[getto(move)-1]) & 0xFF000000)
                    board[boardHash] ^= RandomEnPassant[(getto(move)-8)&7]; 
            }
            else {
                if ( (board[WHITE]&board[PAWN]) & (SetMaskBB[getto(move)+1] | SetMaskBB[getto(move)-1])  & 0xFF00000000)
                    board[boardHash] ^= RandomEnPassant[(getto(move)+8)&7]; 
            }
        }

    }

    // Eval increment    
    if (doscore && USE_INC_EVAL) {
        board[boardScore]+= evalIncrement(board, bindex, som, pto, to);
    }


    board[PEMPTY] = 0;

}

static void undomove(Bitboard *board, Boardindex *bindex, Move move, int som, bool doscore, U64 hash, Cr CR) {

    Square from  = getfrom(move);
    Square to    = getto(move);
    Square cpt   = getcpt(move);

    Piece pfrom  = getpfrom(move);
    Piece pto    = getpto(move);
    Piece pcpt   = getpcpt(move);


    // Eval decrement    
    if (doscore && USE_INC_EVAL) {
        board[boardScore]-= evalIncrement(board, bindex, som, pto, to);
    }


    // Castle move kingside unmove rook
    if ( (pfrom == KING) && to-from == 2 ) {
        /* unset to rook */
        board[som]      &= ClearMaskBB[to-1];
        board[ROOK]     &= ClearMaskBB[to-1];
        bindex[to-1]     = PEMPTY;
        // unset to rook
        if (doscore && USE_INC_EVAL ) {
            board[boardScore]-= evalIncrement(board, bindex, som, ROOK, to-1);
        }

        /* from restore rook */
        board[som]      |= SetMaskBB[from+3];
        board[ROOK]     |= SetMaskBB[from+3];
        bindex[from+3]   = ROOK; 
        // set from rook
        if (doscore && USE_INC_EVAL ) {
            board[boardScore]+= evalIncrement(board, bindex, som, ROOK, from+3);
        }
    }
    // Castle move queenside unmove rook
    else if ( (pfrom == KING) && from-to == 2 ) {
        /* unset to rook */
        board[som]      &= ClearMaskBB[to+1];
        board[ROOK]     &= ClearMaskBB[to+1];
        bindex[to+1]     = PEMPTY;
        // unset to rook
        if (doscore && USE_INC_EVAL ) {
            board[boardScore]-= evalIncrement(board, bindex, som, ROOK, to+1);
        }

        /* from restore rook */
        board[som]      |= SetMaskBB[from-4];
        board[ROOK]     |= SetMaskBB[from-4];
        bindex[from-4]   = ROOK;
        // set from rook
        if (doscore && USE_INC_EVAL ) {
            board[boardScore]+= evalIncrement(board, bindex, som, ROOK, from-4);
        }
    }

    // normal move, including king castle
    // unset to
    board[som]  &= ClearMaskBB[to];
    board[pto]  &= ClearMaskBB[to];
    bindex[to]   = PEMPTY;
    // capture restore
    if (pcpt != PEMPTY ) {
        board[!som] |= SetMaskBB[cpt];
        board[pcpt] |= SetMaskBB[cpt];
        bindex[cpt]  = pcpt;
    }
    // from restore
    board[som]   |= SetMaskBB[from];
    board[pfrom] |= SetMaskBB[from];
    bindex[from]  = pfrom;

    // restore castle rights
    board[boardCR] = getcr(move);

    // restore hash and castle right
    if (doscore) {
        board[boardHash] = hash;
        board[boardCR] = CR;
    }

    // Eval decrement    
    if (doscore && USE_INC_EVAL) {

        board[boardScore]+= evalIncrement(board, bindex, som, pfrom, from);
        if (pcpt >= PAWN && pcpt <= KING) {
            board[boardScore]+= evalIncrement(board, bindex, !som, pcpt, cpt);
        }
    }

    board[PEMPTY] = 0;
}



/* ############################# */
/* ###   piece in check      ### */
/* ############################# */

static int pieceincheck(Bitboard *board, int som, Square checkpos) {

    Bitboard bbWork = 0;
    Bitboard bbMoves = 0;
    Bitboard bbBlockers = 0;

    bbBlockers = board[BLACK] | board[WHITE];

        // Queens
        bbWork = board[som] & board[QUEEN];
        bbMoves = queen_attacks_bb(checkpos, bbBlockers) ;
        if (bbMoves & bbWork)
            return true;
        // Rooks
        bbWork = board[som] & board[ROOK];
        bbMoves = rook_attacks_bb(checkpos, bbBlockers);
        if (bbMoves & bbWork) 
            return true;
        // Bishops
        bbWork = board[som] & board[BISHOP];
        bbMoves = bishop_attacks_bb(checkpos, bbBlockers);
        if (bbMoves & bbWork) 
            return true;
        // Knights
        bbWork = board[som] & board[KNIGHT];
        bbMoves = AttackTablesTo[5][checkpos] ;
        if (bbMoves & bbWork) 
            return true;
        // Pawns
        bbWork = board[som] & board[PAWN];
        bbMoves = som? AttackTablesTo[4][checkpos] : AttackTablesTo[2][checkpos] ;
        if (bbMoves & bbWork) 
            return true;
        // King
        bbWork = board[som] & board[KING];
        bbMoves = AttackTablesTo[6][checkpos] ;
        if (bbMoves & bbWork) 
            return true;

    return false;
}


/* ############################# */
/* ###     move generator    ### */
/* ############################# */

static int genmoves_general(Bitboard *board, Boardindex *bindex, Move *moves, int movecounter, int som, Move lastmove, int qs) {

    Piece p, pto, pcpt;
    Square pos, to, cpt; 
    Move move;
    Bitboard bbTemp = 0;
    Bitboard bbWork = 0;
    Bitboard bbMoves = 0;
    Bitboard bbBlockers = 0;

    bbBlockers = board[BLACK] | board[WHITE];

    // for each piecetype    
    for (p=PAWN;p<=KING;p++) {

        bbWork = board[som] & board[p];  

        while(bbWork) {
            pos = pop_1st_bit(&bbWork);
            bbTemp = 0;
            bbMoves = 0;

            switch(p) {

                case PAWN: {
                    // White Pawn forward steps
                    bbTemp  |= (!qs && !som) ? (AttackTables[1][pos]&(~bbBlockers & SetMaskBB[pos+8]))            : 0 ;
                    // White Pawn double square
                    bbTemp  |= (!qs && !som && ((pos>>3) == 1 ) && (~bbBlockers & SetMaskBB[pos+8]) && (~bbBlockers & SetMaskBB[pos+16]) ) ? SetMaskBB[pos+16] : 0;
                    // White Pawn attacks
                    bbTemp  |= (!som) ? (AttackTables[2][pos] & board[BLACK])   : 0 ;
                    // Black Pawn forward steps
                    bbTemp  |=  (!qs && som) ? (AttackTables[3][pos]&(~bbBlockers & SetMaskBB[pos-8]))            : 0 ;
                    // Black Pawn double square
                    bbTemp  |= (!qs &&  som && ((pos>>3) == 6 ) && (~bbBlockers & SetMaskBB[pos-8]) && (~bbBlockers & SetMaskBB[pos-16]) ) ? SetMaskBB[pos-16] : 0 ;
                    // Black Pawn attacks
                    bbTemp  |= (som) ? (AttackTables[4][pos] & board[WHITE])    : 0 ;
                }
                break;

                case KNIGHT: {
                    bbTemp  = AttackTables[5][pos];
                }
                break;

                case BISHOP: {
                    bbTemp  = bishop_attacks_bb(pos, bbBlockers);
                }
                break;

                case ROOK: {
                    bbTemp  = rook_attacks_bb(pos, bbBlockers);
                }
                break;

                case QUEEN: {
                    bbTemp  = queen_attacks_bb(pos, bbBlockers);
                }
                break;

                case KING: {
                    bbTemp  = AttackTables[6][pos];
                }
                break;
            }

            // captures
            bbMoves  = bbTemp & board[!som];

            // non captures
            bbMoves |= qs? 0 : bbTemp & ~bbBlockers;

            while(bbMoves){

                to = pop_1st_bit(&bbMoves);
                cpt = to;
                pcpt = bindex[cpt];

                // pawn promo always queen
                pto = ( (p == PAWN) && (( !som && (((to>>3)) == 7) ) || ( som && ((to>>3) == 0) )) ) ? QUEEN : p;

                move = makemove(pos, to, cpt, p, pto, pcpt, board[boardCR] & 0xF);

                // evalMove or MVV LVA move ordering 
                move = (pcpt == PEMPTY)? setscore(move, (evalMove(pto, to, som)-evalMove(p, pos, som))) : setscore(move, (EvalPieceValues[pcpt]*16 - EvalPieceValues[pto]));

                domove(board, bindex, move, som, false, 0);
                if (!kingincheck(board, som)) {
                    moves[movecounter] = move;
                    movecounter++;
                }
                undomove(board, bindex, move, som, false,0,0);
            }
        }
    }


    // en passant move
    if (is_pdsq_move(lastmove, !som)) {
        // white
        if (!som) {
            bbTemp = board[PAWN] & board[som] & 0xFF00000000 & (SetMaskBB[getto(lastmove)+1] | SetMaskBB[getto(lastmove)-1]);
        }
        // black     
        else {
            bbTemp = board[PAWN] & board[som] & 0xFF000000 & (SetMaskBB[getto(lastmove)+1] | SetMaskBB[getto(lastmove)-1]);
        }
        while(bbTemp){
            pos = pop_1st_bit(&bbTemp);
            cpt = getto(lastmove);
            to  = som? getto(lastmove)-8 : getto(lastmove)+8;
            pcpt = PAWN;

            move = makemove(pos, to, cpt, PAWN, PAWN, PAWN, board[boardCR] & 0xF);
            move = setscore(move, EvalPieceValues[PAWN]*16 - EvalPieceValues[PAWN]);

            // legal moves only
            domove(board, bindex, move, som, false, 0);
            if (!kingincheck(board, som)) {
                moves[movecounter] = move;
                movecounter++;
            }
            undomove(board, bindex, move, som, false,0,0);
        }
    }
    // gen castle moves
    if (!qs) {
        p = KING;
	    pos = first_1(board[KING] & board[som]);
        if ( checkcastle_kingside(board, som) && ( ( (!som && pos == 4 ) || (som && pos == 60)) ) && ( bindex[pos+1] == PEMPTY  && bindex[pos+2] == PEMPTY && bindex[pos+3] == ROOK) ) {
            if ( !kingincheck(board,som) && !pieceincheck(board, !som, pos+1) && !pieceincheck(board, !som, pos+2) ) {
                move = makemove(pos, pos+2, pos+2, p, p, PEMPTY, board[boardCR] & 0xF);
                move = setscore(move, 10000);

                moves[movecounter] = move;
                movecounter++;
            }
        }
        if ( checkcastle_queenside(board, som) &&  ( ((!som && pos == 4 ) || (som && pos == 60)) && ( bindex[pos-1] == PEMPTY && bindex[pos-2] == PEMPTY && bindex[pos-3] == PEMPTY && bindex[pos-4] == ROOK)) ) {
            if ( !kingincheck(board,som) && !pieceincheck(board, !som, pos-1) && !pieceincheck(board, !som, pos-2) ) {
                move = makemove(pos, pos-2, pos-2, p, p, PEMPTY, board[boardCR] & 0xF);
                move = setscore(move, 10000);

                moves[movecounter] = move;
                movecounter++;
            }
        }
    }
    // sort the moves
    qsort(moves, movecounter, sizeof(Move), cmp_move_desc);

    return movecounter;
}

static int genmoves_attacks(Bitboard *board, Boardindex *bindex, Move *moves, int movecounter, int som, Move lastmove, int ply) {

    Piece p, pto, pcpt;
    Square pos, to, cpt; 
    Move move;
    Bitboard bbTemp = 0;
    Bitboard bbWork = 0;
    Bitboard bbMoves = 0;
    Bitboard bbBlockers = 0;

    bbBlockers = board[BLACK] | board[WHITE];


    // for each piecetype    
    for (p=PAWN;p<=KING;p++) {

        bbWork = board[som] & board[p];  

        while(bbWork) {
            pos = pop_1st_bit(&bbWork);
            bbTemp = 0;
            bbMoves = 0;

            switch(p) {

                case PAWN: {
                    // White Pawn attacks
                    bbTemp  |= (!som) ? (AttackTables[2][pos] & board[BLACK])   : 0 ;
                    // Black Pawn attacks
                    bbTemp  |= (som) ? (AttackTables[4][pos] & board[WHITE])    : 0 ;
                }
                break;

                case KNIGHT: {
                    bbTemp  = AttackTables[5][pos];
                }
                break;

                case BISHOP: {
                    bbTemp  = bishop_attacks_bb(pos, bbBlockers);
                }
                break;

                case ROOK: {
                    bbTemp  = rook_attacks_bb(pos, bbBlockers);
                }
                break;

                case QUEEN: {
                    bbTemp  = queen_attacks_bb(pos, bbBlockers);
                }
                break;

                case KING: {
                    bbTemp  = AttackTables[6][pos];
                }
                break;
            }

            // captures
            bbMoves  = bbTemp & board[!som];

            while(bbMoves){

                to = pop_1st_bit(&bbMoves);
                cpt = to;
                pcpt = bindex[cpt];

                // pawn promo always queen
                pto = ( (p == PAWN) && (( !som && (((to>>3)) == 7) ) || ( som && ((to>>3) == 0) )) ) ? QUEEN : p;

                move = makemove(pos, to, cpt, p, pto, pcpt, board[boardCR] & 0xF);

                // MVVLVA Move ordering
                move = setscore(move, EvalPieceValues[pcpt]*16 - EvalPieceValues[pto]);

                // pseudo legal moves
                moves[movecounter] = move;
                movecounter++;
            }
        }
    }

    // en passant move
    if (is_pdsq_move(lastmove, !som)) {
        // white
        if (!som) {
            bbTemp = board[PAWN] & board[som] & 0xFF00000000 & (SetMaskBB[getto(lastmove)+1] | SetMaskBB[getto(lastmove)-1]);
        }
        // black     
        else {
            bbTemp = board[PAWN] & board[som] & 0xFF000000 & (SetMaskBB[getto(lastmove)+1] | SetMaskBB[getto(lastmove)-1]);
        }
        while(bbTemp){
            pos = pop_1st_bit(&bbTemp);
            cpt = getto(lastmove);
            to  = som? getto(lastmove)-8 : getto(lastmove)+8;
            pcpt = PAWN;

            move = makemove(pos, to, cpt, PAWN, PAWN, PAWN, board[boardCR] & 0xF);
            move = setscore(move, EvalPieceValues[PAWN]*16 - EvalPieceValues[PAWN]);

            // legal moves only
            domove(board, bindex, move, som, false, 0);
            if (!kingincheck(board, som)) {
                moves[movecounter] = move;
                movecounter++;
            }
            undomove(board, bindex, move, som, false,0,0);
        }
    }

    // sort the moves
    qsort(moves, movecounter, sizeof(Move), cmp_move_desc);

    return movecounter;
}

void gen_attack_table(int *attackTableFrom, int *attackTableTo, Bitboard *board, Boardindex *bindex, int som) {

    Piece p, pto, pcpt;
    Square pos, to, cpt; 
    Move move;
    Bitboard bbTemp = 0;
    Bitboard bbWork = 0;
    Bitboard bbMoves = 0;
    Bitboard bbBlockers = 0;

    bbBlockers = board[BLACK] | board[WHITE];

    // for each piecetype    
    for (p=PAWN;p<=KING;p++) {

        bbWork = board[som] & board[p];  

        while(bbWork) {
            pos = pop_1st_bit(&bbWork);
            bbTemp = 0;
            bbMoves = 0;

            switch(p) {

                case PAWN: {
                    // White Pawn attacks
                    bbTemp  |= (!som) ? (AttackTables[2][pos] & board[BLACK])   : 0 ;
                    // Black Pawn attacks
                    bbTemp  |= (som) ? (AttackTables[4][pos] & board[WHITE])    : 0 ;
                }
                break;

                case KNIGHT: {
                    bbTemp  = AttackTables[5][pos];
                }
                break;

                case BISHOP: {
                    bbTemp  = bishop_attacks_bb(pos, bbBlockers);
                }
                break;

                case ROOK: {
                    bbTemp  = rook_attacks_bb(pos, bbBlockers);
                }
                break;

                case QUEEN: {
                    bbTemp  = queen_attacks_bb(pos, bbBlockers);
                }
                break;

                case KING: {
                    bbTemp  = AttackTables[6][pos];
                }
                break;
            }

            // captures
            bbMoves  = bbTemp & board[!som];

            while(bbMoves){

                to = pop_1st_bit(&bbMoves);
                cpt = to;
                pcpt = bindex[cpt];

                // pawn promo always queen
                pto = ( (p == PAWN) && (( !som && (((to>>3)) == 7) ) || ( som && ((to>>3) == 0) )) ) ? QUEEN : p;

                move = makemove(pos, to, cpt, p, pto, pcpt, board[boardCR] & 0xF);

                domove(board, bindex, move, som, false, 0);
                if (!kingincheck(board, som)) {
                    attackTableTo[som*64+cpt]++;
                    attackTableFrom[som*64+pos]++;
                }
                undomove(board, bindex, move, som, false, 0,0);
            }
        }
    }
}



static int genmoves_nonattacks(Bitboard *board, Boardindex *bindex, Move *moves, int movecounter, int som, Move lastmove, int ply) {

    Piece p, pto, pcpt;
    Square pos, to, cpt; 
    Move move;
    Bitboard bbTemp = 0;
    Bitboard bbWork = 0;
    Bitboard bbMoves = 0;
    Bitboard bbBlockers = 0;

    bbBlockers = board[BLACK] | board[WHITE];

    // for each piecetype    
    for (p=PAWN;p<=KING;p++) {

        bbWork = board[som] & board[p];  

        while(bbWork) {
            pos = pop_1st_bit(&bbWork);
            bbTemp = 0;
            bbMoves = 0;

            switch(p) {

                case PAWN: {
                    // White Pawn forward steps
                    bbTemp  |= (!som) ? (AttackTables[1][pos]&(~bbBlockers & SetMaskBB[pos+8]))            : 0 ;
                    // White Pawn double square
                    bbTemp  |= (!som && ((pos>>3) == 1 ) && (~bbBlockers & SetMaskBB[pos+8]) && (~bbBlockers & SetMaskBB[pos+16]) ) ? SetMaskBB[pos+16] : 0;
                    // Black Pawn forward steps
                    bbTemp  |=  (som) ? (AttackTables[3][pos]&(~bbBlockers & SetMaskBB[pos-8]))            : 0 ;
                    // Black Pawn double square
                    bbTemp  |= (som && ((pos>>3) == 6 ) && (~bbBlockers & SetMaskBB[pos-8]) && (~bbBlockers & SetMaskBB[pos-16]) ) ? SetMaskBB[pos-16] : 0 ;
                }
                break;

                case KNIGHT: {
                    bbTemp  = AttackTables[5][pos];
                }
                break;

                case BISHOP: {
                    bbTemp  = bishop_attacks_bb(pos, bbBlockers);
                }
                break;

                case ROOK: {
                    bbTemp  = rook_attacks_bb(pos, bbBlockers);
                }
                break;

                case QUEEN: {
                    bbTemp  = queen_attacks_bb(pos, bbBlockers);
                }
                break;

                case KING: {
                    bbTemp  = AttackTables[6][pos];
                }
                break;
            }

            // non captures
            bbMoves = bbTemp & (~bbBlockers);

            while(bbMoves){

                to = pop_1st_bit(&bbMoves);
                cpt = to;
                pcpt = bindex[cpt];

                // pawn promo always queen
                pto = ( (p == PAWN) && (( !som && (((to>>3)) == 7) ) || ( som && ((to>>3) == 0) )) ) ? QUEEN : p;

                move = makemove(pos, to, cpt, p, pto, pcpt, board[boardCR] & 0xF);

               // set KILLERS score
                if (getmove(move)== getmove(Killers[ply][0])) {
                    move = setscore(move, INF);
                }
                else if (getmove(move)== getmove(Killers[ply][1])) {
                    move = setscore(move, INF-1);
                }
                else if (getmove(move)== getmove(CounterMoves[getfrom(lastmove)][getto(lastmove)]) ) {
                    move = setscore(move, INF-2);
                }
                else {
                    // normal move eval
                    move = setscore(move, evalMove(pto, to, som)-evalMove(p, pos, som));
                }

                // pseudo legal moves
                moves[movecounter] = move;
                movecounter++;
            }
        }
    }

    // gen castle moves
    p = KING;
    pos = first_1(board[p] & board[som]);
    if ( checkcastle_kingside(board, som) && ( ( (!som && pos == 4 ) || (som && pos == 60)) ) && ( bindex[pos+1] == PEMPTY  && bindex[pos+2] == PEMPTY && bindex[pos+3] == ROOK) ) {
        if ( !kingincheck(board,som) && !pieceincheck(board, !som, pos+1) && !pieceincheck(board, !som, pos+2) ) {
            move = makemove(pos, pos+2, pos+2, p, p, PEMPTY, board[boardCR] & 0xF);
            move = setscore(move, 10000);

            moves[movecounter] = move;
            movecounter++;
        }
    }
    if ( checkcastle_queenside(board, som) &&  ( ((!som && pos == 4 ) || (som && pos == 60)) && ( bindex[pos-1] == PEMPTY && bindex[pos-2] == PEMPTY && bindex[pos-3] == PEMPTY && bindex[pos-4] == ROOK)) ) {
        if ( !kingincheck(board,som) && !pieceincheck(board, !som, pos-1) && !pieceincheck(board, !som, pos-2) ) {
            move = makemove(pos, pos-2, pos-2, p, p, PEMPTY, board[boardCR] & 0xF);
            move = setscore(move, 10000);

            moves[movecounter] = move;
            movecounter++;
        }
    }

    // sort the moves
    qsort(moves, movecounter, sizeof(Move), cmp_move_desc);

    return movecounter;
}





/* ############################# */
/* ###      evaluation       ### */
/* ############################# */

void save_killer(Move move, Score score, int ply){

    if(score > getscore(Killers[ply][0]) )
        Killers[ply][0] = setscore(move, score);
    else if(score > getscore(Killers[ply][1]) )
        Killers[ply][1] = setscore(move, score);
}

static int cmp_move_desc(const void *ap, const void *bp) {

    const Move *a = ap;
    const Move *b = bp;

    return getscore(*b) - getscore(*a);
}

int kingincheck(Bitboard *board, int side) {

    Bitboard bbKing = board[KING] & board[side];
    Square kingpos = first_1(bbKing);

    return pieceincheck(board, !side, kingpos);
}

int checkmate(Bitboard *board, Boardindex *bindex, int som, Move lastmove , int kic) {
    
    Move moves[256];
    int movecounter = 0;
    // only if king in check
    if (kic) {
        // call legal move coutner
        movecounter = genmoves_general(board, bindex, moves, 0, som, lastmove, false);
        // no move left, we are in checkmate
        if (movecounter == 0)
            return true;
        else
            return false;
    }
    // we are not in check, so no checkmate
    return false;
}


/* ############################# */
/* ###    searchalgorithm   ### */
/* ############################# */

static Score perft(Bitboard *board, Boardindex *bindex, int som, int depth, Move lastmove ){

    Move moves[256];
    bool kic = false;
    int i = 0;
    int movecounter = 0;
    Score score = 0;
    Cr CR = board[boardCR];
    U64 hash = board[boardHash];

    kic = kingincheck(board, som);

    if (depth == max_depth) {
        NODECOUNT++;
        return 0;
    }

    movecounter = genmoves_general(board, bindex, moves, movecounter, som, lastmove, false);
    MOVECOUNT+= movecounter;

    if (movecounter ==0 && kic) {
        return 0;
    }
    if (movecounter ==0 && !kic) {
        return 0;
    }

    // iterate through moves
    for (i=0;i<movecounter; i++){

        domove(board, bindex, moves[i], som, true, lastmove);

        score = -perft(board, bindex, !som, depth+1, moves[i]);

        undomove(board, bindex, moves[i], som, true, hash, CR);
    }
    return score;
}

static Score alphabeta_qs(Bitboard *board, Boardindex *bindex, int som, int mdepth, int ply, Move lastmove, Score alpha, Score beta, bool nodecount) {

    Move moves[256];
    int i,movecounter = 0;
    int legalmovecounter = 0;
    Score score = alpha;
    int kic = false;
    U64 hash =0;
    Cr CR = board[boardCR];

    hash = USE_INC_HASH ? board[boardHash] : compute_hash(board, som, lastmove);

    // check time
/*
    end = clock();
    elapsed = ((double) (end - start)) / CLOCKS_PER_SEC;
    if ( elapsed >= max_time_per_move ) {
        TIMEOUT = true;
        return 0;
    }
*/

    if (nodecount)
        NODECOUNT++;
    
    // board evaluation
    kic = kingincheck(board, som);

    // call full search
/*
    if (kic) 
        return alphabeta(board, bindex, som, 1, ply, lastmove, alpha, beta, nodecount, true);
*/

    score = som? -board[boardScore] : board[boardScore];

    // lazy eval
    if (USE_LAZY_EVAL && !kic && (score > beta+EvalPieceValues[QUEEN] || score < alpha-EvalPieceValues[QUEEN] )) {
        score = som? -board[boardScore] : board[boardScore];
    }
    else {
        score = som? -eval(board, bindex, som, kic, ply) : eval(board, bindex, som, kic, ply);
    }
    // checkmate
    if (score <= -MATESCORE)
        return score;

 	// Mate Distance Pruning
    if ( USE_MATE_D_PRUNING ) { 
	    alpha = MAX((-INF+ply), alpha);
	    beta  = MIN(-(-INF+ply+1), beta);
	    if (alpha >= beta)
		    return alpha;
    }

    // stand pat
    if (!kic && score >= beta)
        return beta;
    if (!kic && score > alpha)
        alpha = score;


    movecounter = genmoves_attacks(board, bindex, moves, movecounter, som, lastmove, ply);

    MOVECOUNT+= movecounter;

    // iterate through moves
    for (i=0; i < movecounter; i++) {

        if (TIMEOUT == true)
            return 0;

        domove(board, bindex, moves[i], som, true, lastmove);

        if (kingincheck(board, som)) {
            undomove(board, bindex, moves[i], som, true, hash, CR);
            continue;
        }

        legalmovecounter++;

        score = -alphabeta_qs(board, bindex, !som, mdepth-1, ply+1, moves[i], -beta, -alpha, nodecount);
        undomove(board, bindex, moves[i], som, true, hash, CR);

        if (score >= beta) {
            beta_cut++;
            return score;
        }
        if (score > alpha) {
            alpha_cut++;
            alpha = score;
        }
    }

    if (legalmovecounter == 0)
        return score;

    return alpha;
}


static Move iid_search(Bitboard *board, Boardindex *bindex, int som, int mdepth, int ply, Move lastmove, Score alpha, Score beta, bool nodecount) {

    Move bestmove = 0;
    Move moves[256];
    int i = 0;
    int movecounter = 0;
    Score score = alpha;
    U64 hash =0;
    Cr CR = board[boardCR];

    hash = USE_INC_HASH ? board[boardHash] : compute_hash(board, som, lastmove);

    movecounter = genmoves_general(board, bindex, moves, movecounter, som, lastmove, false);

    if (movecounter <= 1) {
        return 0;
    }

    // iterate through moves
    for (i=0;i<movecounter; i++){

        domove(board, bindex, moves[i], som, true, lastmove);

        score = -alphabeta(board, bindex, !som, mdepth-1,  ply+1, moves[i], -beta, -alpha, false, true);

        undomove(board, bindex, moves[i], som, true, hash, CR);

        if (score >= beta) {
            bestmove = moves[i];
            return bestmove;
        }

        if (score > alpha ) {
            alpha = score;
            bestmove = moves[i];
        }

    }
    return bestmove;
}



static Score alphabeta(Bitboard *board, Boardindex *bindex, int som, int mdepth, int ply, Move lastmove, Score alpha, Score beta, bool nodecount, bool nullmove) {

    Move moves_cap[256];
    Move moves_noncap[256];
    Move bestmove = 0;
    Move tempmove = 0;
    Move firstmove = 0;
    Score score = alpha;
    Score boardscore = 0;
    bool pvnode = ((beta - alpha) > 1);
    bool lmr = false;
    signed char flag = FAILLOW;
    int reduction = 0;
    int i = 0;
    int movecounter_cap = 0;
    int movecounter_noncap = 0;
    int legalmovecounter = 0;
    int kic = false;
    int rep_counter = 0;
    int mdepth_org = mdepth;
    int newdepth = mdepth;
    struct tt *tt = NULL;
    U64 hash =0;
    Cr CR = board[boardCR];

    hash = USE_INC_HASH ? board[boardHash] : compute_hash(board, som, lastmove);
    
    // check time
    end = clock();
    elapsed = ((double) (end - start)) / CLOCKS_PER_SEC;
    if ( elapsed >= max_time_per_move ) {
        TIMEOUT = true;
        return 0;
    }

    // check repetition
    Hash_Stack[ply] = hash;
    for (i=ply; i>=0;i-=2) {
        if (Hash_Stack[i] == hash) 
            rep_counter++;
        if (rep_counter == 2)
            return 0;
    }

    kic = kingincheck(board, som);

    // Extend search depth
    if (kic || (getpfrom(lastmove)==PAWN && getpto(lastmove)==QUEEN) )
        mdepth++;

    // call quiescence search
    if (mdepth <= 0) 
        return alphabeta_qs(board, bindex, som, mdepth, ply, lastmove, alpha, beta, nodecount);

    if (nodecount)
        NODECOUNT++;


 	// Mate Distance Pruning
    if ( USE_MATE_D_PRUNING ) { 
	    alpha = MAX((-INF+ply), alpha);
	    beta  = MIN(-(-INF+ply+1), beta);
	    if (alpha >= beta)
		    return alpha;
    }

    // Razoring
    if (USE_RAZOR && mdepth <= 3 && !kic && mdepth == mdepth_org && !pvnode) {
        score = som? -eval(board, bindex, som, kic, ply) : eval(board, bindex, som, kic, ply);
        if (score + EvalPieceValues[QUEEN] <= alpha ) {
            score =  alphabeta_qs(board, bindex, som, 0, ply, lastmove, alpha, beta, nodecount);
            if (score <= alpha)
                return score;
        }
    }

    // check transposition table 
    tt = load_from_tt(hash);

    if (tt != NULL && tt->hash == hash ) {
        // update and check bounds
        if (USE_TT_BOUNDS && tt->depth >= mdepth ) {
            if (tt->flag == EXACTSCORE || tt->flag == FAILHIGH  )  {
                TT_alpha++;
                alpha = MAX(alpha, tt->score);
            }
            if (tt->flag == EXACTSCORE || tt->flag == FAILLOW  )  {
                beta  = MIN(beta, tt->score);
                TT_beta++;
            }
            if (alpha >= beta) return alpha;
        }
        // get TT move
        if ( USE_TT_MOVE && (tt->flag == FAILHIGH || tt->flag == EXACTSCORE ) && (getmove(tt->bestmove)) ) {
            tempmove = getmove(tt->bestmove);
            TT_move++;
        }
    }

    // null move pruning
    reduction = mdepth>6? 4:3;
    if (USE_NULL && lastmove && nullmove && !pvnode && mdepth > 2 && !kic && beta < MATESCORE && alpha < MATESCORE ) {

        donullmove(board, bindex, som, lastmove);
        score = -alphabeta(board, bindex, !som, MAX(mdepth-1-reduction,1), ply+1, 0, -beta, -alpha, nodecount, true);
        undonullmove(board, bindex, som, hash);

        // verification search
        if (score >= beta && score < MATESCORE )
            score = alphabeta(board, bindex, som, MAX(mdepth-1-5,1), ply, lastmove, alpha, beta, nodecount, false);

        if (score >= beta && score < MATESCORE ) {
            save_to_tt(hash, 0, score, FAILHIGH, mdepth, ply, false);
            return score;
        }
    }

    // IID
    if (USE_IID && !tempmove && mdepth > 5 ) {
        tempmove = iid_search(board, bindex, som, mdepth/5,  ply, lastmove, -INF, INF, false);
    }

    // check TT and IID Move
    firstmove = 0;
    if (getmove(tempmove)) {
        firstmove = setcr(tempmove, board[boardCR]&0xF);
    }

    // bestmove first
    if (getmove(firstmove)) {

        domove(board, bindex, firstmove, som, true, lastmove);

        if (!kingincheck(board, som)) {

            score = -alphabeta(board, bindex, !som, mdepth-1,  ply+1, firstmove, -beta, -alpha, nodecount, nullmove);

            legalmovecounter++;

            if (score >= beta) {

                // CounterMove Heuristics
                if (getpcpt(firstmove) == PEMPTY)
                    CounterMoves[getfrom(lastmove)][getto(lastmove)] = getmove(firstmove);
                // Killer Move Heuristics
                if (getpcpt(firstmove) == PEMPTY)
                    save_killer(firstmove, score, ply);

                beta_cut++;
                save_to_tt(hash, firstmove, score, FAILHIGH, mdepth, ply, false);
                undomove(board, bindex, firstmove, som, true, hash, CR);
                return score;
            }
            if (score > alpha) {
                alpha_cut++;
                alpha = score;
                flag = EXACTSCORE;
                bestmove = firstmove;
            }
        }
        undomove(board, bindex, firstmove, som, true, hash, CR);
    }

    movecounter_cap = genmoves_attacks(board, bindex, moves_cap, movecounter_cap, som, lastmove, ply);

    if (nodecount)
        MOVECOUNT+= movecounter_cap;

    // iterate through moves, CAPTURES
    for (i=0; i < movecounter_cap; i++) {

        if (TIMEOUT == true)
            return 0;

        // firstmoves not again
        if (getmove(moves_cap[i]) == getmove(firstmove))
            continue;

        domove(board, bindex, moves_cap[i], som, true, lastmove);

        if (kingincheck(board, som)) {
            undomove(board, bindex, moves_cap[i], som, true, hash, CR);
            continue;
        }

        legalmovecounter++;
    
        newdepth = mdepth;

        if (!pvnode)
            score = -alphabeta(board, bindex, !som, newdepth-1, ply+1, moves_cap[i], -beta, -alpha, nodecount, nullmove);
        else {
            score = -alphabeta(board, bindex, !som, newdepth-1, ply+1, moves_cap[i], -alpha-1, -alpha, nodecount, nullmove);
            if (score > alpha ) {
                score = -alphabeta(board, bindex, !som, newdepth-1, ply+1, moves_cap[i], -beta, -alpha, nodecount, nullmove);
            }
        }

        undomove(board, bindex, moves_cap[i], som, true, hash, CR);

        if (score >= beta) {
            beta_cut++;
            save_to_tt(hash, moves_cap[i], score, FAILHIGH, mdepth, ply, false);
            return score;
        }
        if (score > alpha) {
            alpha_cut++;
            alpha = score;
            flag = EXACTSCORE;
            bestmove = moves_cap[i];
        }
    }


    movecounter_noncap = genmoves_nonattacks(board, bindex, moves_noncap, movecounter_noncap, som, lastmove, ply);

    if (nodecount)
        MOVECOUNT+= movecounter_noncap;

    boardscore = som? -eval(board, bindex, som, kic, ply) : eval(board, bindex, som, kic, ply);

    // iterate through moves, NON-CAPTURES
    for (i=0; i < movecounter_noncap; i++) {

        if (TIMEOUT == true)
            return 0;

        // firstmoves not again
        if (getmove(moves_noncap[i]) == getmove(firstmove))
            continue;

        domove(board, bindex, moves_noncap[i], som, true, lastmove);

        if (kingincheck(board, som)) {
            undomove(board, bindex, moves_noncap[i], som, true, hash, CR);
            continue;
        }

        legalmovecounter++;

        // Futility Pruning
        if ( USE_FUTILITY && mdepth == 1 && !kic && mdepth == mdepth_org && !(getpfrom(moves_noncap[i])==PAWN && getpto(moves_noncap[i])==QUEEN) && boardscore + EvalPieceValues[QUEEN] < alpha ) {
            undomove(board, bindex, moves_noncap[i], som, true, hash, CR);
            continue;
        }

        // Late Move Reductions
        newdepth = mdepth;
        lmr = false;
        if ( USE_LMR && !pvnode && (ply-PLY >= 3) && mdepth > 1 && mdepth <= 4 && !kic && mdepth == mdepth_org && alpha < MATESCORE && (count_1s(board[BLACK])>2 || count_1s(board[WHITE])>2) && !kingincheck(board, !som)) {
            newdepth = (mdepth < 3)? 1 : mdepth -2;
            lmr = true;
        }
        else if ( USE_LMR && legalmovecounter > 0 && !kic && mdepth == mdepth_org ) {
            newdepth = (mdepth < 2)? 1 : mdepth -1;
            lmr = true;
        }
/*
        if ( USE_LMR && !pvnode && (ply-PLY >= 3) && mdepth > 1 && mdepth <= 4 && !kic && mdepth == mdepth_org && alpha < MATESCORE && (count_1s(board[BLACK])>2 || count_1s(board[WHITE])>2) && !kingincheck(board, !som)) {
            newdepth = (mdepth < 3)? 1 : mdepth -2;
            lmr = true;
        }
*/
/*
        else if ( USE_LMR && legalmovecounter > 0 && mdepth <= 4 && !kic && mdepth == mdepth_org ) {
            newdepth = (mdepth < 2)? 1 : mdepth -1;
            lmr = true;
        }
*/
        if (!pvnode)
            score = -alphabeta(board, bindex, !som, newdepth-1, ply+1, moves_noncap[i], -beta, -alpha, nodecount, nullmove);
        else {
            score = -alphabeta(board, bindex, !som, newdepth-1, ply+1, moves_noncap[i], -alpha-1, -alpha, nodecount, nullmove);
            if (score > alpha ) {
                score = -alphabeta(board, bindex, !som, newdepth-1, ply+1, moves_noncap[i], -beta, -alpha, nodecount, nullmove);
            }
        }

        // LMR Research
        if (lmr && score > alpha) {
            newdepth = mdepth_org;
            score = -alphabeta(board, bindex, !som, newdepth-1, ply+1, moves_noncap[i], -beta, -alpha, nodecount, nullmove);
        }

        undomove(board, bindex, moves_noncap[i], som, true, hash, CR);

        if (score >= beta) {
            // CounterMove Heuristics
            if (getpcpt(moves_noncap[i]) == PEMPTY)
                CounterMoves[getfrom(lastmove)][getto(lastmove)] = getmove(moves_noncap[i]);
            // Killer Move Heuristics
            if (getpcpt(moves_noncap[i]) == PEMPTY)
                save_killer(moves_noncap[i], score, ply);

            beta_cut++;
            save_to_tt(hash, moves_noncap[i], score, FAILHIGH, mdepth, ply, false);
            return score;
        }
        if (score > alpha) {
            alpha_cut++;
            alpha = score;
            flag = EXACTSCORE;
            bestmove = moves_noncap[i];
        }
    }


    // stalemate
    if (legalmovecounter == 0 && !kic) {
        return 0;
    }

    // checkmate 
    if (legalmovecounter == 0 && kic) {
        return -INF+ply;
    }

    save_to_tt(hash, bestmove, alpha, flag, mdepth, ply, false);

    return alpha;

}


/* ############################# */
/* ###      root search      ### */
/* ############################# */

static Move rootsearch(Bitboard *board, Boardindex *bindex, int som, int depth, Move lastmove) {

    Move bestmove = 0;
    Move rootmove = 0;
    Move moves[256];
    Move tempmove = 0;
    bool kic = false;
    bool pvsearch = true;
    int i = 0;
    int iid = 0;
    int movecounter = 0;
    Score score = -INF;
    Score tempscore = 0;
    Score alpha = -INF;
    Score beta  =  INF;
    struct tt *tt = NULL;
    U64 hash =0;
    Cr  CR = board[boardCR];

    hash = USE_INC_HASH ? board[boardHash] : compute_hash(board, som, lastmove);

    MOVECOUNT = 0;
    NODECOUNT = 0;
    alpha_cut = 0;
    beta_cut  = 0;
    TT_alpha  = 0;
    TT_beta   = 0;
    TT_move   = 0;
    debug_I   = 0;
    debug_II  = 0;

    // time settings
    start = clock();
    elapsed = 0;
    TIMEOUT = false;

    Hash_Stack[PLY] = hash;

    kic = kingincheck(board, som);

    movecounter = genmoves_general(board, bindex, moves, movecounter, som, lastmove, false);
    MOVECOUNT+= movecounter;

    if (movecounter==0 && kic) {
//        print_checkmate(som);
        return 0;
    }
    if (movecounter==0 && !kic) {
//        print_stalemate();
        return 0;
    }

    // get book move
    if (USE_BOOK) {
        bestmove = book_move(board,bindex,som,lastmove);
        bestmove = setcr(bestmove, board[boardCR] & 0xF);
        for(i=0;i<movecounter;i++) {
            if (getmove(bestmove) == getmove(moves[i])) {
                return bestmove;
            }
        }
        bestmove = 0;
    }

    // check transposition table
    tt = load_from_tt(hash);
    if (USE_TT_MOVE && tt != NULL && tt->hash == hash) {
        if ( tt->flag == FAILHIGH || tt->flag == EXACTSCORE) {
            for(i=0;i<movecounter;i++) {
                if (getmove(tt->bestmove) == getmove(moves[i])) {
                    bestmove = setcr(tt->bestmove, board[boardCR] & 0xF);
                }
            }
        }
    }

    // get a bestomve anyway
    if (!bestmove)
        bestmove = moves[0];

    // get a rootmove anyway
    rootmove = bestmove;


    printf("depth score time nodes move\n");

    while (++iid <= depth && elapsed*2 < max_time_per_move) {

        alpha = -INF;
        beta  =  INF;
        pvsearch = false;

        // PVS 
        domove(board, bindex, bestmove, som, true, lastmove);
        score = -alphabeta(board, bindex, !som, iid-1,  PLY+1, bestmove, -beta, -alpha, true, true);
        undomove(board, bindex, bestmove, som, true, hash, CR);

        tempmove = bestmove;
        tempscore = score;

        if (score > alpha ) {
            alpha = score;
        }

        if (TIMEOUT == true) {
            break;
        }

        // iterate through moves
        for (i=0;i<movecounter; i++){

            // bestmove not again
            if (getmove(moves[i]) == getmove(tempmove) ) {
                moves[i] = setscore(moves[i], tempscore);
                continue;
            }

            domove(board, bindex, moves[i], som, true, lastmove);

            if (pvsearch)
                score = -alphabeta(board, bindex, !som, iid-1,  PLY+1, moves[i], -beta, -alpha, true, true);
            else {
                score = -alphabeta(board, bindex, !som, iid-1,  PLY+1, moves[i], -alpha-1, -alpha, true, true);
                if (score > alpha ) {
 
                    // debug print
                    if (TIMEOUT == false && PRINT_INFO == true) {
                        end = clock();
                        elapsed = ((double) (end - start)) / CLOCKS_PER_SEC;
                        printf("%i %i %i %lu ", iid, alpha, (int)(elapsed * 100), NODECOUNT);          
                        print_debug_movealg(moves[i]);  
                    }

                    score = -alphabeta(board, bindex, !som, iid-1,  PLY+1, moves[i], -beta, -alpha, true, true);
                }
            }

            undomove(board, bindex, moves[i], som, true, hash, CR);

            if (TIMEOUT == true) {
                break;
            }    
    
            if (score > alpha ) {
                moves[i] = setscore(moves[i], score);
                alpha = score;
                bestmove = moves[i];
            }
            else
                moves[i] = setscore(moves[i], score-i);

        }
        
        // set rootmove
        if (TIMEOUT == false) {
            rootmove = bestmove;
        }

        // sort moves 
        qsort(moves, movecounter, sizeof(Move), cmp_move_desc);

        // debug print
        end = clock();
        elapsed = ((double) (end - start)) / CLOCKS_PER_SEC;
        if (TIMEOUT == false && PRINT_INFO == true) {
            printf("%i %i %i %lu ", iid, alpha, (int)(elapsed * 100), NODECOUNT);          
            print_debug_movealg(rootmove);
        }
    }

    end = clock();
    elapsed = ((double) (end - start)) / CLOCKS_PER_SEC;


    fflush(stdout);
    
    best_score = alpha;

    return rootmove;
}




/* ############################# */
/* ### main loop, and Xboard ### */
/* ############################# */
int main(void) {

    char line[256];
    char command[256];
    char c_usermove[256];
    bool go = false;
    Move move;
    Move usermove;
    int i = 0;

    int test_found = 0;
    double test_total_time = 0;
    long test_total_nodes = 0;
    
    setbuf(stdout, NULL);
    setbuf(stdin, NULL);
    setvbuf(stdout, NULL, _IONBF, 0);
    setvbuf(stdin, NULL, _IONBF, 0);   


    srand ( time(NULL) );
    init_bitboards();
    init_tables();
    init_tt();
    if (USE_BOOK)
        book_open();

    printf("\nZetaDva 022 by Srdja Matovic, Montenegro\n");
    printf("Copyright (C) 2011 Srdja Matovic <srdja@matovic.de>\n\n");


	signal(SIGINT, SIG_IGN);

    // main loop
    for(;;) {

        if (!xboard_mode) printf("> ");   
        fflush(stdout);

		if (!fgets(line, 256, stdin)) {
        }

		if (line[0] == '\n') {
			continue;
        }

		sscanf(line, "%s", command);

		if (!strcmp(command, "perft")) {			

	        sscanf(line, "perft %i", &i); 

            max_time_per_move = 1000000;

            NODECOUNT = 0;
            printf("Doing Perft depth %i:\n", i);  
            max_depth = i;
            perft(BOARD, BINDEX, SOM, 0, Lastmove);
            printf("Nodecount:%lu \n", NODECOUNT);

            continue;
        }

        // hash testing
		if (!strcmp(command, "hash")) {			

            PLY = 0;
            SOM = WHITE;
            setboard("setboard rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1");

            Lastmove = 0;
            Hash_Stack[PLY] = USE_INC_HASH ? BOARD[boardHash] : compute_hash(BOARD, SOM, Lastmove);


            sscanf("usermove e2e4", "usermove %s", c_usermove);
            usermove = move_parser(c_usermove, BINDEX, SOM);
            domove(BOARD, BINDEX, usermove, SOM, true, Lastmove);
            PLY++;
            SOM = !SOM;
            print_board(BOARD, BINDEX);
            Lastmove = usermove;
            Hash_Stack[PLY] = USE_INC_HASH ? BOARD[boardHash] : compute_hash(BOARD, SOM, Lastmove);

            sscanf("usermove d7d5", "usermove %s", c_usermove);
            usermove = move_parser(c_usermove, BINDEX, SOM);
            domove(BOARD, BINDEX, usermove, SOM, true, Lastmove);
            PLY++;
            SOM = !SOM;
            print_board(BOARD, BINDEX);
            Lastmove = usermove;
            Hash_Stack[PLY] = USE_INC_HASH ? BOARD[boardHash] : compute_hash(BOARD, SOM, Lastmove);

            sscanf("usermove e4e5", "usermove %s", c_usermove);
            usermove = move_parser(c_usermove, BINDEX, SOM);
            domove(BOARD, BINDEX, usermove, SOM, true, Lastmove);
            PLY++;
            SOM = !SOM;
            print_board(BOARD, BINDEX);
            Lastmove = usermove;
            Hash_Stack[PLY] = USE_INC_HASH ? BOARD[boardHash] : compute_hash(BOARD, SOM, Lastmove);

            sscanf("usermove f7f5", "usermove %s", c_usermove);
            usermove = move_parser(c_usermove, BINDEX, SOM);
            domove(BOARD, BINDEX, usermove, SOM, true, Lastmove);
            PLY++;
            SOM = !SOM;
            print_board(BOARD, BINDEX);
            Lastmove = usermove;
            Hash_Stack[PLY] = USE_INC_HASH ? BOARD[boardHash] : compute_hash(BOARD, SOM, Lastmove);


            sscanf("usermove e1e2", "usermove %s", c_usermove);
            usermove = move_parser(c_usermove, BINDEX, SOM);
            domove(BOARD, BINDEX, usermove, SOM, true, Lastmove);
            PLY++;
            SOM = !SOM;
            print_board(BOARD, BINDEX);
            Lastmove = usermove;
            Hash_Stack[PLY] = USE_INC_HASH ? BOARD[boardHash] : compute_hash(BOARD, SOM, Lastmove);

            sscanf("usermove e8f7", "usermove %s", c_usermove);
            usermove = move_parser(c_usermove, BINDEX, SOM);
            domove(BOARD, BINDEX, usermove, SOM, true, Lastmove);
            PLY++;
            SOM = !SOM;
            print_board(BOARD, BINDEX);
            Lastmove = usermove;
            Hash_Stack[PLY] = USE_INC_HASH ? BOARD[boardHash] : compute_hash(BOARD, SOM, Lastmove);

            for(i=0;i<=PLY;i++) {
                printf("# hash %llx \n", Hash_Stack[i]);
            }
        }    


		if (!strcmp(command, "test")) {			


            char filename[256];
            char bm[5];
            char temp[255];
            char movec[6];
            char testline [256];
            int i = 0;

	        sscanf(line, "test %s", filename); 

            FILE *file = fopen ( filename, "r" );

            PRINT_INFO = true;
            max_depth = 9;
            max_time_per_move = 1000000;
            test_found = 0;
            test_total_time = 0;
            test_total_nodes = 0;
            
            if ( file != NULL ) {
                 
                while ( fgets ( testline, sizeof(testline), file ) != NULL ) {

                    i++;
                    init_tt();
        	        sscanf(testline, "%s %s %s %s bm %[0-9a-zA-Z]s", temp, temp, temp, temp, bm); 
                    setboard_epd(testline);
        
                    printf("Doing EPD #: %i\n", i);

                    PLY =0;
                    move = rootsearch(BOARD, BINDEX, SOM, max_depth, Lastmove);

                    printf("\n");
                    printf("searched: %s\n", bm);
                    move2alg(move, movec);
                    if ( strstr(movec, bm))  {
                        test_found++;
                    }
                    printf("found:    %s\n", movec);

                    printf("time: %f \n",elapsed);
                    test_total_time += elapsed;
                    test_total_nodes += NODECOUNT;

                }
                fclose ( file );
            }


            printf("\n\n");
            printf("found %i from %i: \n", test_found, i);
            printf("total seconds: %i \n", (int)test_total_time);
            printf("total nodes: %li \n", test_total_nodes);
            printf("\n\n");

            free(TT);
            if (USE_BOOK)
                book_close();
            exit(0);

        }

		if (!strcmp(command, "xboard")) {			
            xboard_mode = true;
            printf("feature done=0\n");  
            continue;
        }
        if (strstr(command, "protover")) {
		    printf("feature myname=\"ZetaDva022\"\n");
            printf("feature reuse=1\n");
            printf("feature setboard=1\n");
            printf("feature memory=1\n");
            printf("feature smp=1\n");
            printf("feature usermove=1\n");
            printf("feature san=0\n");
            printf("feature time=1\n");
            printf("feature debug=1\n");
            printf("feature done=1\n");
		    continue;
        }        
        if (!strcmp (command, "time")) {
	        sscanf(line, "time %lf", &time_left_computer); 
	        max_time_per_move = ( (time_left_computer - (time_per_move*100)) / ( max_moves- ( ((PLY+1)/2) % max_moves)) ) /100;
            max_time_per_move = max_time_per_move <= 0 ? 1 : max_time_per_move;
	        continue;
        }    
        if (!strcmp(command, "otim")) { 
	        sscanf(line, "otim %lf", &time_left_opponent); 
	        continue;
        }	
        if (!strcmp (command, "level")){
            sscanf(line, "level %i %s", &max_moves, time_string);
            if (!strcmp (time_string, ":")) {
                sscanf(time_string, "%i:%i", &time_minutes, &time_seconds);
            }
            else {
                sscanf(time_string, "%i", &time_minutes);
                time_seconds = 0;
            }
            max_moves = (max_moves == 0)? 40 : max_moves;
	        max_time_per_move = (((time_minutes*60) + time_seconds) / max_moves);
	        time_per_move = (((time_minutes*60) + time_seconds) / max_moves);
            continue;
        }

		if (!strcmp(command, "new")) {
            setboard("setboard rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1");
            init_tt();
            Lastmove = 0;
            PLY =0;
            go = false;
            force_mode = false;
            random_mode = false;
			continue;
		}
		if (!strcmp(command, "setboard")) {
            setboard(line);
            init_tt();
            Lastmove = 0;
            PLY =0;
            go = false;
            force_mode = false;
            random_mode = false;
			continue;
		}         	
		if (!strcmp(command, "quit")) {
            if (TT != NULL)
                free(TT);
            if (USE_BOOK)
                book_close();
			exit(0);
        }
        // TODO: randomize eval
		if (!strcmp(command, "random")) {
            random_mode = random_mode? false :true;
			continue;
		}
		if (!strcmp(command, "force")) {
            force_mode = true;
			continue;
		}
		if (!strcmp(command, "white")) {
			continue;
		}
		if (!strcmp(command, "black")) {
			continue;
		}
		if (!strcmp(command, "sd")) {
			sscanf(line, "sd %i", &max_depth);
			continue;
		}
		if (!strcmp(command, "memory")) {
			sscanf(line, "memory %i", &max_mem_mb);
            init_tt();
			continue;
		}
		if (!strcmp(command, "cores")) {
			sscanf(line, "cores %i", &max_cores);
			continue;
		}
        if (!strcmp(command, "go")) {
            force_mode = false;
            if (!go) {
                Hash_Stack[PLY] = compute_hash(BOARD, SOM, Lastmove);
                move = rootsearch(BOARD, BINDEX, SOM, max_depth, Lastmove);
                if (move != 0) { 
                    Lastmove = move;
                    PLY++;
                    domove(BOARD, BINDEX, move, SOM, true, Lastmove);
                    print_movealg(move);
                    SOM = !SOM;
                }
            }
            continue;
        }		
		if (!strcmp(command, "result")) {
           char result[256];
           sscanf(line,"result %s",result);
           if (best_score > 29900) 
           printf("\ncomputer mates\n"); // send mate result to winboard - JA
           continue;
        }
        if (strstr(command, "usermove")){
            sscanf(line, "usermove %s", c_usermove);
            Hash_Stack[PLY] = compute_hash(BOARD, SOM, Lastmove);
            usermove = move_parser(c_usermove, BINDEX, SOM);
            Lastmove = usermove;
            domove(BOARD, BINDEX, usermove, SOM, true, Lastmove);
            PLY++;
            print_board(BOARD, BINDEX);
            SOM = !SOM;
            if (!force_mode) {
                go = true;
                Hash_Stack[PLY] = compute_hash(BOARD, SOM, Lastmove);
                move = rootsearch(BOARD, BINDEX, SOM, max_depth, Lastmove);
                if (move != 0) { 
                    domove(BOARD, BINDEX, move, SOM, true, Lastmove);
                    PLY++;
                    Lastmove = move;
                    print_board(BOARD, BINDEX);
                    print_movealg(move);
                    printf("\neval score = %i\n",best_score);
                    SOM = !SOM;
                }
            }
            continue;
        }   
    }

   return 0;
}


/* ############################# */
/* ###        parser         ### */
/* ############################# */

Move move_parser(char *usermove, Boardindex *bindex, int som) {

    int file;
    int rank;
    Square from,to,cpt;
    Piece pto, pfrom, pcpt;
    Move move;
    char promopiece;

    file = (int)usermove[0] -97;
    rank = (int)usermove[1] -49;
    from = make_square(file,rank);
    file = (int)usermove[2] -97;
    rank = (int)usermove[3] -49;
    to = make_square(file,rank);

    pfrom = bindex[from];
    pto = bindex[from];
    cpt = to;

    /* en passant */
    cpt = ( pfrom == PAWN &&  som && (from>>3) == 3 && from-to != 8 && bindex[to] == PEMPTY ) ? to+8 : cpt;
    cpt = ( pfrom == PAWN && !som && (from>>3) == 4 && to-from != 8 && bindex[to] == PEMPTY ) ? to-8 : cpt;
    pcpt = bindex[cpt];

    /* pawn promo piece */
    promopiece = usermove[4];
    if (promopiece == 'q' || promopiece == 'Q' )
        pto = QUEEN;
    else if (promopiece == 'n' || promopiece == 'N' )
        pto = KNIGHT;
    else if (promopiece == 'b' || promopiece == 'B' )
        pto = BISHOP;
    else if (promopiece == 'r' || promopiece == 'R' )
        pto = ROOK;

    move = makemove(from, to, cpt, pfrom, pto , pcpt, BOARD[boardCR] & 0xF);

    return move;
}

static void setboard(char *fen) {

    int i, j, side, kic =0;
    Square ep_sq;
    int index;
    int file = 0;
    int rank = 7;
    int pos  = 0;
    char temp;
    char position[255];
    char csom;
    char castle[5];
    char ep[3];
    int bla;
    int blubb;
    char string[26] = {"  PNBRQK  pnbrqk/12345678"};
    U64 hash = 0;    

	sscanf(fen, "setboard %s %c %s %s %i %i", position, &csom, castle, ep, &bla, &blubb);
		
    for(i=0;i<64;i++) {
        BINDEX[i] = PEMPTY;
    }
    BOARD[boardCR] = 0;
    for(i=0;i<12;i++) {
        BOARD[i] = 0;
    }

    i =0;
    while (!(rank==0 && file==8)) {
        temp = position[i];
        i++;        
        for (j=0;j<25;j++) {
    		if (temp == string[j]) {
                if (j == 16) {
                    rank--;
                    file=0;
                }
                else if (j >=17) {
                    file+=j-16;
                }
                else {
                    pos = (rank*8) + file;
                    side = (j>7) ? 1 :0;
                    index = side? j-8 : j;
                    BOARD[side]  |= SetMaskBB[pos];
                    BOARD[index] |= SetMaskBB[pos];
                    BINDEX[pos] = index;
                    file++;
                }
                break;                
            }
        }
    }

    // site on move
    if (csom == 'w' || csom == 'W') {
        SOM = WHITE;
    }
    if (csom == 'b' || csom == 'B') {
        SOM = BLACK;
    }

    // Castle Rights
    if (castle[0] != '-') {
    
        // white kingside
        if (castle[0] == 'K')
            setcastle_kingside(BOARD, WHITE);
        // white queenside
        if (castle[1] == 'Q')
            setcastle_queenside(BOARD, WHITE);
        // black queenside
        if (castle[3] == 'q')
            setcastle_queenside(BOARD, BLACK);
        // black kingside
        if (castle[2] == 'k')
            setcastle_kingside(BOARD, BLACK);
    }

printf("\n checkcastle_queenside:%i\n", checkcastle_queenside(BOARD, SOM));
printf("\n checkcastle_queenside:%i\n", checkcastle_queenside(BOARD, !SOM));

    // compute hash
    hash =  book_gen_key(BOARD, SOM, 0);

    // add en passant
    if (ep[0] != '-') {
        file = (int)ep[0] -97;
        rank = (int)ep[1] -49;
        ep_sq = make_square(file,rank);

        if (SOM) {
            if (0xFF000000 & BOARD[BLACK]&BOARD[PAWN] & (SetMaskBB[ep_sq+8+1] | SetMaskBB[ep_sq+8-1] )) {
                hash ^= RandomEnPassant[ep_sq&7];
            }
        }
        else {
            if ( 0xFF00000000 & BOARD[WHITE]&BOARD[PAWN] & (SetMaskBB[ep_sq-8-1] | SetMaskBB[ep_sq-8+1] ) ) {
                hash ^= RandomEnPassant[ep_sq&7];
            }
        }
    }

    // init hash
    Hash_Stack[PLY]  = hash;
    BOARD[boardHash] = hash;

    // init eval
    kic = kingincheck(BOARD, SOM);
    BOARD[boardScore] = eval(BOARD, BINDEX, SOM, kic, PLY);

    print_board(BOARD,BINDEX);

}


static void setboard_epd(char *epd) {

    int i, j, side, kic =0;
    Square ep_sq;
    int index;
    int file = 0;
    int rank = 7;
    int pos  = 0;
    char temp;
    char position[255];
    char csom;
    char castle[5];
    char ep[3];
    char string[26] = {"  PNBRQK  pnbrqk/12345678"};
    U64 hash = 0;    

	sscanf(epd, "%s %c %s %s ", position, &csom, castle, ep);
		
    for(i=0;i<64;i++) {
        BINDEX[i] = PEMPTY;
    }
    for(i=0;i<12;i++) {
        BOARD[i] = 0;
    }

    i =0;
    while (!(rank==0 && file==8)) {
        temp = position[i];
        i++;        
        for (j=0;j<25;j++) {
    		if (temp == string[j]) {
                if (j == 16) {
                    rank--;
                    file=0;
                }
                else if (j >=17) {
                    file+=j-16;
                }
                else {
                    pos = (rank*8) + file;
                    side = (j>7) ? 1 :0;
                    index = side? j-8 : j;
                    BOARD[side]  |= SetMaskBB[pos];
                    BOARD[index] |= SetMaskBB[pos];
                    BINDEX[pos] = index;
                    file++;
                }
                break;                
            }
        }
    }

    // Castle Rights
    if (castle[0] != '-') {
    
        if (castle[0] == 'K')
            setcastle_kingside(BOARD, WHITE);
        else
            clearcastle_kingside(BOARD, WHITE);
        if (castle[1] == 'Q')
            setcastle_queenside(BOARD, WHITE);
        else
            clearcastle_queenside(BOARD, WHITE);
        if (castle[2] == 'k')
            setcastle_kingside(BOARD, BLACK);
        else
            clearcastle_kingside(BOARD, BLACK);
        if (castle[3] == 'q')
            setcastle_queenside(BOARD, BLACK);
        else
            clearcastle_queenside(BOARD, BLACK);
    }
    else {
        clearcastle_kingside(BOARD, WHITE);
        clearcastle_queenside(BOARD, WHITE);
        clearcastle_kingside(BOARD, BLACK);
        clearcastle_queenside(BOARD, BLACK);
    }     

    // site on move
    if (csom == 'w' || csom == 'W') {
        SOM = WHITE;
    }
    if (csom == 'b' || csom == 'B') {
        SOM = BLACK;
    }

    // compute hash
    hash =  book_gen_key(BOARD, SOM, 0);

    // add en passant
    if (ep[0] != '-') {
        file = (int)ep[0] -97;
        rank = (int)ep[1] -49;
        ep_sq = make_square(file,rank);

        if (SOM) {
            if (0xFF000000 & BOARD[BLACK]&BOARD[PAWN] & (SetMaskBB[ep_sq+8+1] | SetMaskBB[ep_sq+8-1] )) {
                hash ^= RandomEnPassant[ep_sq&7];
            }
        }
        else {
            if ( 0xFF00000000 & BOARD[WHITE]&BOARD[PAWN] & (SetMaskBB[ep_sq-8-1] | SetMaskBB[ep_sq-8+1] ) ) {
                hash ^= RandomEnPassant[ep_sq&7];
            }
        }
    }

    // init hash
    Hash_Stack[PLY]  = hash;
    BOARD[boardHash] = hash;

    // init eval
    kic = kingincheck(BOARD, SOM);
    BOARD[boardScore] = eval(BOARD, BINDEX, SOM, kic, PLY);

    print_board(BOARD,BINDEX);
}

/* ############################# */
/* ###       print tools     ### */
/* ############################# */

void print_movealg(Move move) {


    char movec[6];
    move2alg(move, movec);

    printf("move %s\n", movec);
    fflush(stdout);

}

void move2alg(Move move, char * movec) {

    char rankc[8] = "12345678";
    char filec[8] = "abcdefgh";
    Square from = getfrom(move);
    Square to   = getto(move);
    Piece pfrom   = getpfrom(move);
    Piece pto   = getpto(move);


    movec[0] = filec[square_file(from)];
    movec[1] = rankc[square_rank(from)];
    movec[2] = filec[square_file(to)];
    movec[3] = rankc[square_rank(to)];
    movec[4] = ' ';
    movec[5] = '\0';

    /* pawn promo */
    if ( (pfrom == PAWN && ( ( (to&56)/8 == 7) || ((to&56)/8 == 0) ) )) {
        if (pto == QUEEN)
            movec[4] = 'q';
        if (pto == ROOK)
            movec[4] = 'r';
        if (pto == BISHOP)
            movec[4] = 'b';
        if (pto == KNIGHT)
            movec[4] = 'n';
    }

}


static void print_debug_movealg(Move move) {

    char rankc[8] = "12345678";
    char filec[8] = "abcdefgh";
    char movec[6] = "";
    Square from = getfrom(move);
    Square to   = getto(move);
    Piece pfrom   = getpfrom(move);
    Piece pto   = getpto(move);

    movec[0] = filec[square_file(from)];
    movec[1] = rankc[square_rank(from)];
    movec[2] = filec[square_file(to)];
    movec[3] = rankc[square_rank(to)];

    /* pawn promo */
    if ( (pfrom == PAWN && ( ( (to&56)/8 == 7) || ((to&56)/8 == 0) ) )) {
        if (pto == QUEEN)
            movec[4] = 'q';
        if (pto == ROOK)
            movec[4] = 'r';
        if (pto == BISHOP)
            movec[4] = 'b';
        if (pto == KNIGHT)
            movec[4] = 'n';
    }

    printf("%s\n", movec);
    fflush(stdout);
}

static void print_bitboard(Bitboard board) {

    int i,j,pos;
    printf("###ABCDEFGH###\n");
   
    for(i=8;i>0;i--) {
        printf("#%i ",i);
        for(j=0;j<8;j++) {
            pos = ((i-1)*8) + j;
            if (bit_is_set(board, pos)) 
                printf("x");
            else 
                printf("-");

        }
       printf("\n");
    }
    printf("###ABCDEFGH###\n");
    fflush(stdout);
}

static void print_board(Bitboard *board, Boardindex *bindex) {

    int i,j,pos;
    char wpchars[9] = "  PNBRQK-";
    char bpchars[9] = "  pnbrqk-";

    printf("###ABCDEFGH###\n");
    for(i=8;i>0;i--) {
        printf("#%i ",i);
        for(j=0;j<8;j++) {
            pos = ((i-1)*8) + j;
            if (board[0] & SetMaskBB[pos] )
                printf("%c", wpchars[bindex[pos]]);
            else if (board[1] & SetMaskBB[pos] )
                printf("%c", bpchars[bindex[pos]]);
            else 
                printf("-");
       }
       printf("\n");
    }
    fflush(stdout);
}

static void print_checkmate(int side) {

  if (side == WHITE)
    printf("0-1 {Black mates}");
  if (side == BLACK) 
    printf("1-0 {White mates}");
}

static void print_stalemate() {
    printf("1/2-1/2 {Stalemate}");
}


static void print_stats() {
    FILE 	*Stats;
    Stats = fopen("zeta_dva.dbg", "ab+");
    fprintf(Stats, "nodes: %lu ,moves: %lu ,beta-cuts: %i ,alpha-cuts: %i, TT_MOVE: %i ,TT_BETA: %i ,TT_ALPHA: %i , debug I: %i , debug II: %i ,sec: %f \n", NODECOUNT, MOVECOUNT,beta_cut, alpha_cut, TT_move,  TT_beta, TT_alpha, debug_I, debug_II , elapsed);
    fclose(Stats);
}

